/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// r_surf.c: surface-related refresh code

#include "quakedef.h"
#include "d3d_quake.h"

// number of verts
int d3d_VertexBufferVerts = 0;

// draw in 2 streams
LPDIRECT3DVERTEXBUFFER9 d3d_BrushModelVerts = NULL;

int r_renderflags = 0;

void R_PushDlights (mnode_t *headnode);
void EmitSkyPolys (msurface_t *surf);
void R_DrawSkyChain (msurface_t *skychain);
bool R_CullBox (vec3_t mins, vec3_t maxs);
void D3D_RotateForEntity (entity_t *e, bool shadow = false);
void R_StoreEfrags (efrag_t **ppefrag);
void D3D_MakeLightmapTexCoords (msurface_t *surf, float *v, float *st);
void D3D_GetLightUpdateTimes (void);
void D3D_DrawWorld (void);
void D3D_AddInlineBModelsToTextureChains (void);

// For gl_texsort 0
msurface_t  *skychain = NULL;


cvar_t r_lockpvs ("r_lockpvs", "0");

/*
===============
R_TextureAnimation

Returns the proper texture for a given time and base texture
===============
*/
texture_t *R_TextureAnimationInner (texture_t *base)
{
	int		relative;
	int		count;

	if (currententity->frame)
	{
		if (base->alternate_anims)
			base = base->alternate_anims;
	}

	if (!base->anim_total) return base;

	relative = (int)(cl.time*10) % base->anim_total;

	count = 0;	

	while (base->anim_min > relative || base->anim_max <= relative)
	{
		base = base->anim_next;

		// prevent crash here
		if (!base) return NULL;
		if (++count > 100) return NULL;
	}

	return base;
}


texture_t *R_TextureAnimation (texture_t *base)
{
	// check animation
	texture_t *tex = R_TextureAnimationInner (base);

	// catch errors
	if (!tex) tex = base;

	// update renderflags
	if (tex->d3d_Fullbright)
		r_renderflags |= R_RENDERLUMA;
	else r_renderflags |= R_RENDERNOLUMA;

	// return what we got
	return tex;
}


/*
=============================================================

	WORLD MODEL

=============================================================
*/

/*
================
R_RecursiveWorldNode
================
*/
void R_RecursiveWorldNode (mnode_t *node)
{
	int			c, side;
	mplane_t	*plane;
	msurface_t	*surf, **mark;
	mleaf_t		*pleaf;
	double		dot;

	if (node->contents == CONTENTS_SOLID) return;
	if (node->visframe != r_visframecount) return;
	if (R_CullBox (node->minmaxs, node->minmaxs + 3)) return;

	// if a leaf node, draw stuff
	if (node->contents < 0)
	{
		pleaf = (mleaf_t *)node;

		mark = pleaf->firstmarksurface;
		c = pleaf->nummarksurfaces;

		if (c)
		{
			do
			{
				(*mark)->visframe = r_framecount;
				mark++;
			} while (--c);
		}

		// deal with model fragments in this leaf
		if (pleaf->efrags) R_StoreEfrags (&pleaf->efrags);

		return;
	}

	// node is just a decision point, so go down the apropriate sides
	// find which side of the node we are on
	plane = node->plane;

	switch (plane->type)
	{
	case PLANE_X:
		dot = modelorg[0] - plane->dist;
		break;
	case PLANE_Y:
		dot = modelorg[1] - plane->dist;
		break;
	case PLANE_Z:
		dot = modelorg[2] - plane->dist;
		break;
	default:
		dot = DotProduct (modelorg, plane->normal) - plane->dist;
		break;
	}

	if (dot >= 0)
		side = 0;
	else side = 1;

	// recurse down the children, front side first
	R_RecursiveWorldNode (node->children[side]);

	// add stuff to the draw lists
	c = node->numsurfaces;

	if (c)
	{
		surf = cl.worldbrush->surfaces + node->firstsurface;

		if (dot < 0 -BACKFACE_EPSILON)
			side = SURF_PLANEBACK;
		else if (dot > BACKFACE_EPSILON)
			side = 0;

		for (; c; c--, surf++)
		{
			if (surf->visframe != r_framecount) continue;
			if ((dot < 0) ^ !!(surf->flags & SURF_PLANEBACK)) continue;

			// get the correct animation sequence
			texture_t *tex = R_TextureAnimation (surf->texinfo->texture);

			// if sorting by texture, just store it out
			if (surf->flags & SURF_DRAWSKY)
			{
				// link it in
				surf->texturechain = skychain;
				skychain = surf;
			}
			else if (surf->flags & SURF_DRAWTURB)
			{
				// link it in
				surf->texturechain = tex->texturechain;
				tex->texturechain = surf;

				// mark the texture as having been visible
				tex->visframe = r_framecount;

				// flag for rendering
				r_renderflags |= R_RENDERWATERSURFACE;
			}
			else
			{
				// check under/over water
				if (surf->flags & SURF_UNDERWATER)
					r_renderflags |= R_RENDERUNDERWATER;
				else r_renderflags |= R_RENDERABOVEWATER;

				// check for lightmap modifications
				D3D_CheckLightmapModification (surf);

				// link it in
				surf->texturechain = tex->texturechain;
				tex->texturechain = surf;
			}
		}
	}

	// recurse down the back side
	R_RecursiveWorldNode (node->children[!side]);
}



/*
=============
R_DrawWorld
=============
*/
void R_DrawWorld (void)
{
	entity_t	ent;

	memset (&ent, 0, sizeof(ent));
	ent.model = cl.worldmodel;

	VectorCopy (r_refdef.vieworg, modelorg);

	currententity = &ent;
	currenttexture = -1;

	// build light update times
	D3D_GetLightUpdateTimes ();

	// calculate dynamic lighting for the world
	R_PushDlights (cl.worldbrush->nodes);

	// build the world
	R_RecursiveWorldNode (cl.worldbrush->nodes);

	// add inline models
	D3D_AddInlineBModelsToTextureChains ();

	// unlock all lightmaps that were previously locked
	D3D_UnlockLightmaps ();

	// draw sky first (required by z-buffer mangling going on in there)
	R_DrawSkyChain (skychain);

	D3D_DrawWorld ();
}


/*
===============
R_MarkLeaves
===============
*/
void R_LeafVisibility (byte *vis)
{
	mnode_t *node;

	for (int i = 0; i < cl.worldbrush->numleafs; i++)
	{
		if (vis[i >> 3] & (1 << (i & 7)))
		{
			node = (mnode_t *) &cl.worldbrush->leafs[i + 1];

			do
			{
				// already added
				if (node->visframe == r_visframecount) break;

				// add it
				node->visframe = r_visframecount;
				node = node->parent;
			} while (node);
		}
	}
}


void R_MarkLeaves (void)
{
	byte	*vis;
	mnode_t	*node;
	int		i;
	extern	byte *mod_novis;

	// viewleaf hasn't changed
	if ((r_oldviewleaf == r_viewleaf && !r_novis.value) || r_lockpvs.value) return;

	// go to a new visframe
	r_visframecount++;

	if (r_novis.value)
	{
		// add them all
		R_LeafVisibility (mod_novis);

		// done
		return;
	}

	// add viewleaf to visible lists
	R_LeafVisibility (Mod_LeafPVS (r_viewleaf, cl.worldmodel));

	// start of map
	if (!r_oldviewleaf) return;

	// add old viewleaf on a contents transition
	if (r_viewleaf->contents != r_oldviewleaf->contents)
	{
		// just making sure this gets called when it should (and doesn't when it shouldn't)
		Con_DPrintf ("Adding r_oldviewleaf for contents transition\n");
		R_LeafVisibility (Mod_LeafPVS (r_oldviewleaf, cl.worldmodel));
	}
}


void D3D_BuildWarpSurfVB (msurface_t *surf, model_t *mod, worldvert_t *vdest)
{
	int			i;
	int			lindex;
	float		*vec;

	for (i = 0; i < surf->numedges; i++, vdest++)
	{
		lindex = mod->bh->surfedges[surf->firstedge + i];

		if (lindex > 0)
			vec = mod->bh->vertexes[mod->bh->edges[lindex].v[0]].position;
		else vec = mod->bh->vertexes[mod->bh->edges[-lindex].v[1]].position;

		// verts
		vdest->xyz[0] = vec[0];
		vdest->xyz[1] = vec[1];
		vdest->xyz[2] = vec[2];

		// texcoords
		vdest->st[0] = DotProduct (vec, surf->texinfo->vecs[0]) / 4;
		vdest->st[1] = DotProduct (vec, surf->texinfo->vecs[1]) / 4;

		// store a quarter again in lm texcoords to speed up the pixel shader
		vdest->lm[0] = vdest->st[0] / 4;
		vdest->lm[1] = vdest->st[1] / 4;
	}
}


void D3D_BuildSurfaceVertexBuffer (msurface_t *surf, model_t *mod, worldvert_t *vdest)
{
	int			i, lindex, lnumverts;
	medge_t		*pedges, *r_pedge;
	float		*vec;
	float		s, t;
	mvertex_t	*vertbase = mod->bh->vertexes;

	// reconstruct the polygon
	pedges = mod->bh->edges;
	lnumverts = surf->numedges;

	// go through all of the verts
	for (i = 0; i < surf->numedges; i++, vdest++)
	{
		// build vertexes
		lindex = mod->bh->surfedges[surf->firstedge + i];

		if (lindex > 0)
			vec = mod->bh->vertexes[mod->bh->edges[lindex].v[0]].position;
		else vec = mod->bh->vertexes[mod->bh->edges[-lindex].v[1]].position;

		// copy verts in
		vdest->xyz[0] = vec[0];
		vdest->xyz[1] = vec[1];
		vdest->xyz[2] = vec[2];

		// build texcoords
		s = DotProduct (vec, surf->texinfo->vecs[0]) + surf->texinfo->vecs[0][3];
		s /= surf->texinfo->texture->width;

		t = DotProduct (vec, surf->texinfo->vecs[1]) + surf->texinfo->vecs[1][3];
		t /= surf->texinfo->texture->height;

		// copy them in
		vdest->st[0] = s;
		vdest->st[1] = t;

		// make lightmap coords
		D3D_MakeLightmapTexCoords (surf, vec, vdest->lm);
	}
}


void D3D_PutSurfacesInVertexBuffer (void)
{
	// release it if it already exists
	SAFE_RELEASE (d3d_BrushModelVerts);

	// create our new vertex buffer
	HRESULT hr = d3d_Device->CreateVertexBuffer
	(
		d3d_VertexBufferVerts * sizeof (worldvert_t),
		D3DUSAGE_WRITEONLY,
		(D3DFVF_XYZ | D3DFVF_TEX2),
		D3DPOOL_MANAGED,
		&d3d_BrushModelVerts,
		NULL
	);

	// fixme - fall back on DrawPrimitiveUP if this happens!
	if (FAILED (hr)) Sys_Error ("D3D_PutSurfacesInVertexBuffer: d3d_Device->CreateVertexBuffer failed");

	worldvert_t *verts;

	// lock the full buffers
	d3d_BrushModelVerts->Lock (0, 0, (void **) &verts, 0);

	int vboffset = 0;

	for (int j = 1; j < MAX_MODELS; j++)
	{
		model_t *m = cl.model_precache[j];

		if (!m) break;
		if (m->name[0] == '*') continue;
		if (m->type != mod_brush) continue;

		msurface_t *surf = m->bh->surfaces;

		for (int i = 0; i < m->bh->numsurfaces; i++, surf++)
		{
			// skip
			if (surf->flags & SURF_DRAWSKY) continue;

			// build the vertex buffer for this surf
			if (surf->flags & SURF_DRAWTURB)
				D3D_BuildWarpSurfVB (surf, m, verts);
			else D3D_BuildSurfaceVertexBuffer (surf, m, verts);

			// set offset
			surf->vboffset = vboffset;

			// advance pointers
			verts += surf->numedges;

			// advance offset counter
			vboffset += surf->numedges;
		}
	}

	// unlock and preload, baby!
	d3d_BrushModelVerts->Unlock ();
	d3d_BrushModelVerts->PreLoad ();
}


/*
==================
GL_BuildLightmaps

Builds the lightmap texture
with all the surfaces from all brush models
==================
*/
void GL_BuildLightmaps (void)
{
	int		i, j;
	model_t	*m;

	r_framecount = 1;		// no dlightcache

	// clear down previous lightmaps
	D3D_ReleaseLightmaps ();

	// no verts yet
	d3d_VertexBufferVerts = 0;

	for (j = 1; j < MAX_MODELS; j++)
	{
		m = cl.model_precache[j];

		if (!m) break;
		if (m->name[0] == '*') continue;
		if (m->type != mod_brush) continue;

		msurface_t *surf = m->bh->surfaces;

		for (int i = 0; i < m->bh->numsurfaces; i++, surf++)
		{
			// increment vertex count
			d3d_VertexBufferVerts += surf->numedges;

			// skip these
			if (surf->flags & SURF_DRAWTURB) continue;
			if (surf->flags & SURF_DRAWSKY) continue;

			// create the lightmap
			D3D_CreateSurfaceLightmap (surf);
	
			// no polys on these surfs
			surf->polys = NULL;
		}
	}

	// put all the surfaces into vertex buffers
	D3D_PutSurfacesInVertexBuffer ();

	// upload all lightmaps
	D3D_UploadLightmaps ();
}

