
#include "quakedef.h"
#include "d3d_quake.h"
#include "d3d_hlsl.h"

/*
=============================================================

  SPRITE MODELS

=============================================================
*/

typedef struct sprverts_s
{
	float x, y, z;
	float s, t;
} sprverts_t;

sprverts_t sprverts[4];

/*
================
R_GetSpriteFrame
================
*/
mspriteframe_t *R_GetSpriteFrame (entity_t *currententity)
{
	msprite_t		*psprite;
	mspritegroup_t	*pspritegroup;
	mspriteframe_t	*pspriteframe;
	int				i, numframes, frame;
	float			*pintervals, fullinterval, targettime, time;

	psprite = currententity->model->sh;
	frame = currententity->frame;

	if ((frame >= psprite->numframes) || (frame < 0))
	{
		Con_Printf ("R_DrawSprite: no such frame %d\n", frame);
		frame = 0;
	}

	if (psprite->frames[frame].type == SPR_SINGLE)
	{
		pspriteframe = psprite->frames[frame].frameptr;
	}
	else
	{
		pspritegroup = (mspritegroup_t *)psprite->frames[frame].frameptr;
		pintervals = pspritegroup->intervals;
		numframes = pspritegroup->numframes;
		fullinterval = pintervals[numframes-1];

		time = cl.time + currententity->syncbase;

		// when loading in Mod_LoadSpriteGroup, we guaranteed all interval values
		// are positive, so we don't have to worry about division by 0
		targettime = time - ((int)(time / fullinterval)) * fullinterval;

		for (i=0 ; i<(numframes-1) ; i++)
		{
			if (pintervals[i] > targettime)
				break;
		}

		pspriteframe = pspritegroup->frames[i];
	}

	return pspriteframe;
}


/*
=================
R_DrawSpriteModel

=================
*/
#define VectorScalarMult(a,b,c) {(c)[0] = (a)[0] * (b); (c)[1] = (a)[1] * (b); (c)[2] = (a)[2] * (b);}

void R_DrawSpriteModel (entity_t *e)
{
	vec3_t	point;
	mspriteframe_t	*frame;
	float		*up, *right;
	vec3_t		v_forward, v_right, v_up;
	msprite_t	*psprite;
	vec3_t		fixed_origin;
	vec3_t		temp;

	// don't even bother culling, because it's just a single
	// polygon without a surface cache
	frame = R_GetSpriteFrame (e);
	psprite = currententity->model->sh;

	VectorCopy (e->origin, fixed_origin);

	if (psprite->type == SPR_ORIENTED)
	{
		// bullet marks on walls
		AngleVectors (currententity->angles, v_forward, v_right, v_up);

		VectorScalarMult (v_forward, -2, temp);
		VectorAdd (temp, fixed_origin, fixed_origin);

		up = v_up;
		right = v_right;
	}
	else
	{
		// normal sprite
		up = vup;
		right = vright;
	}

	D3D_BindTexture ((LPDIRECT3DTEXTURE9) frame->texture);

	VectorMA (fixed_origin, frame->down, up, point);
	VectorMA (point, frame->left, right, point);
	sprverts[0].x = point[0]; sprverts[0].y = point[1]; sprverts[0].z = point[2]; sprverts[0].s = 0; sprverts[0].t = 1;

	VectorMA (fixed_origin, frame->up, up, point);
	VectorMA (point, frame->left, right, point);
	sprverts[1].x = point[0]; sprverts[1].y = point[1]; sprverts[1].z = point[2]; sprverts[1].s = 0; sprverts[1].t = 0;

	VectorMA (fixed_origin, frame->up, up, point);
	VectorMA (point, frame->right, right, point);
	sprverts[2].x = point[0]; sprverts[2].y = point[1]; sprverts[2].z = point[2]; sprverts[2].s = 1; sprverts[2].t = 0;

	VectorMA (fixed_origin, frame->down, up, point);
	VectorMA (point, frame->right, right, point);
	sprverts[3].x = point[0]; sprverts[3].y = point[1]; sprverts[3].z = point[2]; sprverts[3].s = 1; sprverts[3].t = 1;

	d3d_Device->DrawPrimitiveUP (D3DPT_TRIANGLEFAN, 2, sprverts, sizeof (sprverts_t));
}


void D3D_DrawSpriteModels (void)
{
	if (!r_drawentities.value) return;
	if (!(r_renderflags & R_RENDERSPRITE)) return;

	d3d_Device->SetVertexDeclaration (d3d_V3ST2Declaration);

	HRESULT hr = d3d_SpriteFX->SetTechnique ("Sprite");

	if (FAILED (hr))
	{
		// spam the console for development purposes but don't crash the engine, just don't draw
		Con_Printf ("Shader Error in Effect \"Sprite.fx\" - technique \"Sprite\" not found\n");
		return;
	}

	hr = d3d_SpriteFX->SetMatrix ("WorldProjMatrix", &(d3d_WorldMatrix * d3d_PerspectiveMatrix));

	if (FAILED (hr))
	{
		// spam the console for development purposes but don't crash the engine, just don't draw
		Con_Printf ("Shader Error in technique \"Sprite\" - Parameter \"WorldProjMatrix\" not found\n");
		return;
	}

	int NumPasses;

	// note - validate that there are two passes in here
	d3d_SpriteFX->Begin ((UINT *) &NumPasses, D3DXFX_DONOTSAVESAMPLERSTATE);

	if (NumPasses != 1)
	{
		// spam the console for development purposes but don't crash the engine, just don't draw
		Con_Printf ("Shader Error in technique \"Sprite\" - NumPasses != 1\n");
		d3d_SpriteFX->End ();
		return;
	}

	d3d_EnableAlphaBlend->Apply ();
	d3d_Device->SetRenderState (D3DRS_ZWRITEENABLE, FALSE);
	d3d_SpriteFX->BeginPass (0);

	for (int i = 0; i < cl_numvisedicts; i++)
	{
		currententity = cl_visedicts[i];

		// not a sprite model
		if (currententity->model->type != mod_sprite) continue;

		// draw it
		R_DrawSpriteModel (currententity);
	}

	d3d_SpriteFX->EndPass ();
	d3d_SpriteFX->End ();
	d3d_Device->SetRenderState (D3DRS_ZWRITEENABLE, TRUE);
	d3d_DisableAlphaBlend->Apply ();
}

